"""Functions for running Clustal Omega"""

from phamerate import subprocess

VERSION = "1.2.4"


class ClustalOmegaError(Exception):
    """Error caused by running Clustal Omega."""
    pass


def get_clustalo_version():
    """Return the Clustal Omega version string.

    :return: version
    """
    command = "clustalo --version"
    try:
        stdout, stderr = subprocess.run(command)
        if stderr:
            print(stderr)
            raise ClustalOmegaError(f"command failed: {command}")
        return stdout.rstrip()
    except FileNotFoundError:
        return None


def version_ok():
    """Check that local Clustal Omega version matches the expected
    version.

    :return: ok
    """
    ok, clustalo_version = True, get_clustalo_version()

    if not clustalo_version:
        ok = False
    elif clustalo_version != VERSION:
        ok = False

    return ok


def run_clustalo(infile, outfile, threads=1, verbose=False):
    """Use clustalo to generate an MSA from the sequences in `infile`
    and store the results in `outfile`.

    :param infile: path to a FASTA multiple sequence file
    :type infile: pathlib.Path
    :param outfile: path to a FASTA multiple sequence alignment file
    :type outfile: pathlib.Path
    :param threads: number of threads to use
    :type threads: int
    :param verbose: return stdout
    :type verbose: bool
    :return: infile, outfile
    """
    command = f"clustalo -i {infile} --infmt=fasta -o {outfile} " \
              f"--outfmt=fasta --output-order=tree-order " \
              f"--threads={threads} --seqtype=protein --force"

    stdout, stderr = subprocess.run(command)
    if verbose and stdout:
        print(stdout)
    if stderr and not stderr.startswith("WARNING: "):
        print(stderr)
        raise ClustalOmegaError(f"command failed: {command}")

    return infile, outfile



"""Functions for reading and writing FASTA files."""


def read_fasta(filepath):
    """Read a FASTA file and return the headers and sequences.

    :param filepath: a FASTA file to parse
    :type filepath: pathlib.Path or str
    :return: sequences
    """
    headers, sequences = list(), list()

    cache = list()
    with open(filepath, "r") as fasta_reader:
        for line in fasta_reader:
            # If the line is a header line, flush the cache and store new header
            if line.startswith(">"):
                sequences.append("".join(cache))
                cache = list()
                headers.append(line.lstrip(">").rstrip())
            # Otherwise, append to the cache
            else:
                cache.append(line.rstrip())

        # Flush the last sequence out of the cache, and pop the empty sequence
        sequences.append("".join(cache))
        sequences = sequences[1:]

    return zip(headers, sequences)


def write_fasta(sequences, filepath, wrap=80):
    """Write the sequences to the indicated file in FASTA format,
    wrapping long sequence lines at the indicated width.

    Use wrap=None or wrap=0 or wrap=False to turn off line-wrapping.

    :param sequences: the headers and translations to write_fasta
    :type sequences: list[tuple[str, str]]
    :param filepath: a FASTA file to write_fasta to
    :type filepath: pathlib.Path or str
    :param wrap: maximum number of characters per sequence line
    :type wrap: int
    :return: filepath
    """
    with open(filepath, "w") as fasta_writer:
        for header, sequence in sequences:
            fasta_writer.write(f">{header}\n")

            # If no wrapping, write_fasta the whole sequence
            if not wrap:
                fasta_writer.write(f"{sequence}\n")
            # Else, use string slicing to split the sequence to satisfy wrap
            else:
                for i in range(0, len(sequence), wrap):
                    fasta_writer.write(f"{sequence[i:i + wrap]}\n")

    return filepath



"""Function for reading Genbank flatfiles."""

from Bio import SeqIO
from Bio.Data.CodonTable import TranslationError


def read_genbank(filepath):
    """Read a Genbank flatfile and return headers and sequences
    for all CDS features.

    :param filepath: the path to a Genbank flatfile
    :type filepath: pathlib.Path or str
    :return: headers, sequences
    """
    headers, sequences = list(), list()

    for record in SeqIO.parse(filepath, "genbank"):
        features = [x for x in record.features if x.type == "CDS"]

        for i, cds in enumerate(features):
            header = f"lcl|{record.id}_"

            # try to get the "protein_id" qualifier, otherwise skip
            protein_id = cds.qualifiers.get("protein_id", [None])[0]
            if protein_id:
                header += f"prot_{protein_id}_"

            # append CDS index + 1
            header += f"{i + 1}"

            # try to get the "gene" qualifier, otherwise skip
            gene = cds.qualifiers.get("gene", [None])[0]
            if gene:
                header += f" [gene={gene}]"

            # try to get the "locus_tag" qualifier, otherwise skip
            locus_tag = cds.qualifiers.get("locus_tag", [None])[0]
            if locus_tag:
                header += f" [locus_tag={locus_tag}]"

            # try to get the "product" qualifier, otherwise skip
            product = cds.qualifiers.get("product", [None])[0]
            if product:
                header += f" [protein={product}]"
            else:
                header += " [protein=hypothetical protein]"

            # append the protein_id
            if protein_id:
                header += f" [protein_id={protein_id}]"

            # append the location
            start, end = cds.location.start, cds.location.end
            if cds.location.strand == 1:
                header += f" [location={start}..{end}]"
            else:
                header += f" [location=complement({start}..{end})]"

            # append the genbank key
            header += f" [gbkey=CDS]"

            # try to get the translation or generate it - skip if cannot
            try:
                translation = str(cds.qualifiers.get("translation", [None])[0])
                if not translation:
                    translation = str(cds.translate(record.seq))
            except TranslationError:
                # Translations encoded by frameshifts or those that use
                # suppressor tRNAs may fail to translate - skip these
                print(f"Unable to translate {header} - skipping it")
                continue

            headers.append(header)
            sequences.append(translation)

    return zip(headers, sequences)



"""Assort phage protein sequences into phamilies of homologs using MMseqs2."""

import argparse
import datetime
import pathlib

from phamerate.multiprocess import CPUS

DATE = datetime.datetime.now().strftime("%d_%b_%Y")
OUT_DIR = pathlib.Path().cwd().joinpath(f"phamerate__{DATE}")

# Sequence-sequence
SEQ_MIN_SEQ_ID = 35.0
SEQ_COVERAGE = 80.0
SEQ_EVALUE = 0.001
SEQ_SENSITIVITY = 7
SEQ_CLUSTER_MODE = 0
SEQ_CLUSTER_STEPS = 1

HMM_MIN_SEQ_ID = 15.0
HMM_COVERAGE = 70.0
HMM_EVALUE = 0.001
HMM_SENSITIVITY = 7
HMM_CLUSTER_MODE = 0
HMM_CLUSTER_STEPS = 3

EPILOG = """
Steinegger M. and Söding J. MMseqs2 enables sensitive protein
sequence searching for the analysis of massive data sets. Nature
Biotechnology, 2017. doi: 10.1038/nbt.3988"""


def parse_args():
    """Parse command line arguments."""
    p = argparse.ArgumentParser(description=__doc__, prog="phamerate",
                                epilog=EPILOG)

    p.add_argument("infile", nargs="+", type=pathlib.Path,
                   help="path to input file(s) in FASTA or Genbank flatfile "
                        "format")

    s = p.add_argument_group("MMseqs2 sequence-sequence clustering arguments")
    s.add_argument("--identity",
                   type=float, default=SEQ_MIN_SEQ_ID, metavar='',
                   help=f"percent identity for sequence-sequence clustering "
                        f"[default: {SEQ_MIN_SEQ_ID}%%]")
    s.add_argument("--coverage",
                   type=float, default=SEQ_COVERAGE, metavar='',
                   help=f"percent coverage for sequence-sequence clustering "
                        f"[default: {SEQ_COVERAGE}%%]")
    s.add_argument("--evalue",
                   type=float, default=SEQ_EVALUE, metavar='',
                   help=f"E-value threshold for sequence-sequence clustering "
                        f"[default: {SEQ_EVALUE}]")
    s.add_argument("--sensitivity",
                   type=float, default=SEQ_SENSITIVITY, metavar='',
                   help=f"sensitivity: 1 favors speed, 7 favors "
                        f"sensitivity [default: {SEQ_SENSITIVITY}]")
    s.add_argument("--cluster-mode",
                   type=int, default=SEQ_CLUSTER_MODE, metavar='',
                   help=f"clustering algorithm [default: {SEQ_CLUSTER_MODE}]")
    s.add_argument("--cluster-steps",
                   type=int, default=SEQ_CLUSTER_STEPS, metavar='',
                   help=f"number of steps for sequence-sequence clustering "
                        f"to proceed in [default: {SEQ_CLUSTER_STEPS}]")

    h = p.add_argument_group("MMseqs2 profile-sequence clustering arguments")
    h.add_argument("--hmm-identity",
                   type=float, default=HMM_MIN_SEQ_ID, metavar='',
                   help=f"percent identity for profile-consensus clustering "
                        f"[default: {HMM_MIN_SEQ_ID}%%]")
    h.add_argument("--hmm-coverage",
                   type=float, default=HMM_COVERAGE, metavar='',
                   help=f"percent coverage for profile-consensus clustering "
                        f"[default: {HMM_COVERAGE}%%]")
    h.add_argument("--hmm-evalue",
                   type=float, default=HMM_EVALUE, metavar='',
                   help=f"E-value threshold for profile-consensus clustering "
                        f"[default: {HMM_EVALUE}]")
    h.add_argument("--hmm-sensitivity",
                   type=float, default=HMM_SENSITIVITY, metavar='',
                   help=f"sensitivity: 1 favors speed, 7 favors "
                        f"sensitivity [default: {HMM_SENSITIVITY}]")
    h.add_argument("--hmm-cluster-mode",
                   type=int, default=HMM_CLUSTER_MODE, metavar='',
                   help=f"clustering algorithm [default: {HMM_CLUSTER_MODE}]")
    h.add_argument("--hmm-cluster-steps",
                   type=int, default=HMM_CLUSTER_STEPS, metavar='',
                   help=f"number of steps for profile-consensus clustering "
                        f"to proceed in [default: {HMM_CLUSTER_STEPS}]")
    h.add_argument("--skip-hmm", action="store_true",
                   help="do not perform profile-consensus clustering")

    p.add_argument("-v", "--verbose", action="store_true",
                   help="print progress messages to the console")
    p.add_argument("-d", "--debug", action="store_true",
                   help="run in debug mode")
    p.add_argument("-a", "--align-phams", action="store_true",
                   help="use Clustal Omega to align phams (could take "
                        "awhile...)")
    p.add_argument("-p", "--pangenome", action="store_true",
                   help="pangenome analysis à la Roary (only meaningful if "
                        "given one input file per genome)")
    p.add_argument("-c", "--cpus", type=int, default=CPUS,
                   help=f"number of threads to use [default: {CPUS}]")
    p.add_argument("-o", "--outdir", default=OUT_DIR, type=pathlib.Path,
                   help=f"path to directory where output files should go "
                        f"[default: {OUT_DIR}]")

    return p.parse_args()



from phamerate.fileio.fasta import read_fasta, write_fasta
from phamerate.fileio.genbank import read_genbank


def sniff_format(filepath):
    """Sniff the first line of the file to determine the file format.

    Return options are: "fasta", "genbank", or None

    :param filepath: the path to the file we need to sniff
    :type filepath: pathlib.Path
    :return: fmt
    """
    with open(filepath, "r") as file_sniffer:
        line = file_sniffer.readline()

    if line.startswith(">"):
        return "fasta"
    elif line.startswith("LOCUS"):
        return "genbank"
    else:
        return None


__all__ = ["read_fasta", "read_genbank", "sniff_format", "write_fasta"]



"""Functions for performing pham assembly with MMseqs2."""

from phamerate import subprocess

VERSION = "13.45111"


class MMseqs2Error(Exception):
    """Error caused by running MMseqs2."""
    pass


def get_mmseqs_version():
    """Return the MMseqs2 version string.

    :return: version
    """
    command = "mmseqs version"

    try:
        stdout, stderr = subprocess.run(command)
        if stderr:
            print(stderr)
            raise MMseqs2Error(f"command failed: {command}")
        return stdout.rstrip()
    except FileNotFoundError:
        return None


def version_ok():
    """Check that local MMseqs2 version matches the expected version.

    :return: ok
    """
    ok, mmseqs_version = True, get_mmseqs_version()

    if not mmseqs_version:
        ok = False
    elif mmseqs_version != VERSION:
        ok = False

    return ok


def createdb(fasta, seq_db, debug=False):
    """Run 'mmseqs createdb' command.

    :param fasta: path to the FASTA file to convert
    :type fasta: pathlib.Path
    :param seq_db: path to the sequence database to create
    :type seq_db: pathlib.Path
    :param debug: print verbose output to console
    :type debug: bool
    """
    command = f"mmseqs createdb {fasta} {seq_db} -v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return seq_db


def cluster(seq_db, clu_db, tmp_dir, identity, coverage, e_value, sens,
            max_seqs=1000, cluster_mode=0, cluster_reassign=True,
            cluster_steps=1, alignment_mode=3, coverage_mode=0, cpus=1,
            debug=False):
    """Run 'mmseqs cluster' command.

    Arguments are mapped to the following MMseqs2 arguments:

    * identity = `--min-seq-id`
    * coverage = `-c`
    * e_value = `-e`
    * sens = `-s`
    * max_seqs = `--max-seqs`
    * cluster_mode = `--cluster-mode`
    * cluster_reassign = `--cluster-reassign`
    * cluster_steps = `--cluster-steps`
    * alignment_mode = `--alignment-mode`
    * coverage_mode = `--cov-mode`
    * cpus = `--threads`

    :param seq_db: path to the sequence database to cluster
    :type seq_db: pathlib.Path
    :param clu_db: path to write_fasta the clustered sequence database
    :type clu_db: pathlib.Path
    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param identity: percent identity threshold for clustering
    :type identity: float
    :param coverage: percent coverage threshold for clustering
    :type coverage: float
    :param e_value: significance threshold for clustering
    :type e_value: float
    :param sens: sensitivity parameter for seeding alignments
    :type sens: float
    :param max_seqs: number of hits to store per query
    :type max_seqs: int
    :param cluster_mode: clustering algorithm to use
    :type cluster_mode: int
    :param cluster_reassign: reassign cluster members that violate
        thresholds
    :type cluster_reassign: bool
    :param cluster_steps: number of clustering steps to proceed with
    :type cluster_steps: int
    :param alignment_mode: what kind of alignments to generate
    :type alignment_mode: int
    :param coverage_mode: coverage calculation to use
    :type coverage_mode: int
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param debug: print verbose output to console
    :type debug: bool
    """
    if identity > 1:
        identity /= 100.0
    if coverage > 1:
        coverage /= 100.0

    command = f"mmseqs cluster {seq_db} {clu_db} {tmp_dir} " \
              f"--min-seq-id {identity} -c {coverage} -e {e_value} " \
              f"-s {sens} --max-seqs {max_seqs} --cluster-mode " \
              f"{cluster_mode} --cluster-steps {cluster_steps} " \
              f"--alignment-mode {alignment_mode} --cov-mode " \
              f"{coverage_mode}  --threads {cpus} -v 3 "

    if cluster_reassign:
        command += " --cluster-reassign"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return clu_db


def result2profile(seq_db, clu_db, pro_db, cpus=1, debug=False):
    """Run 'mmseqs result2profile' command.

    :param seq_db: path to the sequence database to cluster
    :type seq_db: pathlib.Path
    :param clu_db: path to the clustered sequence database
    :type clu_db: pathlib.Path
    :param pro_db: path to the desired profile database
    :type pro_db: pathlib.Path
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param debug: print verbose output to console
    :type debug: bool
    """
    command = f"mmseqs result2profile {seq_db} {seq_db} {clu_db} {pro_db} " \
              f"--threads {cpus} -v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return pro_db


def profile2consensus(pro_db, con_db, cpus=1, debug=False):
    """Run 'mmseqs profile2consensus' command.

    :param pro_db: path to a profile database
    :type pro_db: pathlib.Path
    :param con_db: path to the desired consensus sequence database
    :type con_db: pathlib.Path
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param debug: print verbose output to console
    :type debug: bool
    """
    command = f"mmseqs profile2consensus {pro_db} {con_db} " \
              f"--threads {cpus} -v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return con_db


def search(pro_db, con_db, aln_db, tmp_dir, identity, coverage, e_value, sens,
           max_seqs=1000, cluster_steps=1, alignment_mode=3, coverage_mode=0,
           cpus=1, debug=False):
    """Run 'mmseqs search' command.

    Arguments are mapped to the following MMseqs2 arguments:

    * identity = `--min-seq-id`
    * coverage = `-c`, `--cov`
    * e_value = `-e`, `--e-profile`
    * sens = `-s`
    * max_seqs = `--max-seqs`
    * cluster_steps = `--cluster-steps`
    * alignment_mode = `--alignment-mode`
    * coverage_mode = `--cov-mode`
    * cpus = `--threads`

    :param pro_db: path to a profile database
    :type pro_db: pathlib.Path
    :param con_db: path to a consensus sequence database
    :type con_db: pathlib.Path
    :param aln_db: path to the desired alignment database
    :type aln_db: pathlib.Path
    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param identity: percent identity threshold
    :type identity: float
    :param coverage: percent coverage threshold
    :type coverage: float
    :param e_value: significance threshold
    :type e_value: float
    :param sens: sensitivity parameter for seeding alignments
    :type sens: float
    :param max_seqs: maximum number of hits to store per query
    :type max_seqs: int
    :param cluster_steps: num of iterative profile-consensus searches
        to perform
    :type cluster_steps: int
    :param alignment_mode: MMseqs2 alignment mode
    :type alignment_mode: int
    :param coverage_mode: MMseqs2 coverage mode
    :type coverage_mode: int
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param debug: print verbose output to console
    :type debug: bool
    """
    if identity > 1:
        identity /= 100.0
    if coverage > 1:
        coverage /= 100.0

    command = f"mmseqs search {pro_db} {con_db} {aln_db} {tmp_dir} " \
              f"--min-seq-id {identity} -c {coverage} --cov {coverage} " \
              f"-e {e_value} --e-profile {e_value} -s {sens} --max-seqs " \
              f"{max_seqs} --num-iterations {cluster_steps} " \
              f"--alignment-mode {alignment_mode} --cov-mode " \
              f"{coverage_mode} --add-self-matches --threads {cpus} -v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return aln_db


def clust(con_db, aln_db, res_db, cluster_mode, cpus=1, debug=False):
    """Run 'mmseqs clust' command.

    :param con_db: directory MMseqs2 can use as scratch space
    :type con_db: pathlib.Path
    :param aln_db: directory MMseqs2 can use as scratch space
    :type aln_db: pathlib.Path
    :param res_db: directory MMseqs2 can use as scratch space
    :type res_db: pathlib.Path
    :param cluster_mode: clustering algorithm to use
    :type cluster_mode: int
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param debug: print verbose output to console
    :type debug: bool
    """
    command = f"mmseqs clust {con_db} {aln_db} {res_db} " \
              f"--cluster-mode {cluster_mode} --threads {cpus} -v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return res_db


def createseqfiledb(seq_db, clu_db, sf_db, cpus=1, debug=False):
    """Run 'mmseqs createseqfiledb' command.

    :param seq_db: path to the sequence database
    :type seq_db: pathlib.Path
    :param clu_db: path to the clustered sequence database
    :type clu_db: pathlib.Path
    :param sf_db: path where the seqfiledb should be written
    :type sf_db: pathlib.Path
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param debug: print verbose output to the console
    :type debug: bool
    :return: sf_db
    """
    command = f"mmseqs createseqfiledb {seq_db} {clu_db} {sf_db} " \
              f"--threads {cpus} -v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return sf_db


def result2flat(query_db, target_db, sf_db, outfile, debug=False):
    """Run 'mmseqs result2flat' command after clustering.

    :param query_db: the query database used for clustering
    :type query_db: pathlib.Path
    :param target_db: the target database used for clustering
    :type target_db: pathlib.Path
    :param sf_db: the seqfiledb created after clustering
    :type sf_db: pathlib.Path
    :param outfile: path where the FASTA-like cluster output should go
    :type outfile: pathlib.Path
    :param debug: print verbose output to the console
    :type debug: bool
    :return: outfile
    """
    command = f"mmseqs result2flat {query_db} {target_db} {sf_db} {outfile} " \
              f"-v 3"

    stdout, stderr = subprocess.run(command)
    if debug and stdout:
        print(stdout)
    if stderr:
        print(stderr)
        raise MMseqs2Error(f"command failed: {command}")

    return outfile


def parse_output(outfile):
    """Parse the indicated MMseqs2 FASTA-like file into a dictionary of
    integer-named phams.

    :param outfile: FASTA-like parseable output
    :type outfile: pathlib.Path
    :return: phams
    :rtype: dict
    """
    phams = {}
    pham_translations = list()
    pham_name = -1

    with open(f"{outfile}", "r") as fh:
        prior = fh.readline()
        current = fh.readline()

        # While loop to iterate until EOF
        while current:
            # If current and previous lines are header lines, new pham block
            if current.startswith(">") and prior.startswith(">"):
                phams[pham_name] = pham_translations
                pham_name += 1
                pham_translations = []

            # If the current line is a translation line, keep growing pham
            if not current.startswith(">"):
                pham_translations.append(current.rstrip())

            prior, current = current, fh.readline()

        # Need to write_fasta the last pham into the dictionary
        phams[pham_name] = pham_translations

    # Pham -1 is a placeholder
    phams.pop(-1)

    return phams



"""Classes and functions to perform basic pangenome analysis."""

from phamerate.SequenceDB import SequenceDB, Pham
from phamerate import fileio

summary_template = """
Core genes (99% <= strains <= 100%):    {}
Soft core genes (95% <= strains < 99%): {}
Shell genes (15% <= strains < 95%):     {}
Cloud genes (0% <= strains < 15%):      {}
Total genes (0% <= strains <= 100%):    {}
"""

ROARY_HEADER = ["Gene", "Non unique Gene name", "No. isolates", "No. sequences",
                "Avg sequences per isolate", "Genome Fragment",
                "Order within Fragment", "Accessory Fragment",
                "Accessory Order with Fragment", "QC", "Min group size nuc",
                "Max group size nuc", "Avg group size nuc"]


class Genome(SequenceDB):
    """A class derived from SequenceDB that is better suited for the
    kinds of operations one might wish to do with smaller groups of
    sequences found in the same genome.
    """
    def __init__(self, sequences=None):
        super().__init__(sequences)
        self._pham_geneids = dict()

    def add_phams(self, phams):
        geneids = set(self.geneids)
        for i, pham in enumerate(phams):
            phamid = f"pham_{i+1}"
            shared = geneids.intersection(pham.geneids)
            if len(shared) > 0:
                self._pham_geneids[phamid] = list(shared)

    def __contains__(self, item):
        """Query a genome for whether it contains a pham."""
        return item in self._pham_geneids.keys()

    @property
    def pham_genes(self):
        for phamid, geneids in self._pham_geneids.items():
            yield phamid, geneids


def build_genomes(phams, infiles):
    """Parse input files into genomes, and match them to the phams
    their genes belong to.

    :param phams:
    :type phams: list[Pham]
    :param infiles:
    :type infiles: list[pathlib.Path]
    :return: temp_genomes
    """
    temp_genomes = dict()
    for infile in infiles:
        name = infile.stem

        fmt = fileio.sniff_format(infile)
        if fmt == "fasta":
            sequences = fileio.read_fasta(infile)
        elif fmt == "genbank":
            sequences = fileio.read_genbank(infile)
        else:
            continue

        genome = Genome(sequences)
        genome.add_phams(phams)

        temp_genomes[name] = genome

    return temp_genomes


def analyze_pangenome(phams, infiles, outdir):
    """

    :param phams:
    :type phams: list[Pham]
    :param infiles:
    :type infiles: list[pathlib.Path]
    :param outdir:
    :type outdir: pathlib.Path
    :return:
    """
    # Build genomes to analyze
    genomes = build_genomes(phams, infiles)

    # Map phams to genomes to geneids
    pham_data = dict()
    for name, genome in genomes.items():
        for phamid, geneids in genome.pham_genes:
            if phamid in pham_data:
                pham_data[phamid][name] = geneids
            else:
                pham_data[phamid] = {name: geneids}

    # Analyze pangenome
    core, core_file = list(), outdir.joinpath("core_genes.txt")
    soft, soft_file = list(), outdir.joinpath("soft_core_genes.txt")
    shell, shell_file = list(), outdir.joinpath("shell_genes.txt")
    cloud, cloud_file = list(), outdir.joinpath("cloud_genes.txt")
    strain_rows, roary_rows = list(), list()

    for i, pham in enumerate(phams):
        phamid = f"pham_{i+1}"
        pham_genomes = pham_data[phamid]

        # Begin building this pham's Roary row
        seq_per_iso = float(len(pham)) / len(pham_genomes)
        roary_row = [phamid, "", str(len(pham_genomes)), str(len(pham)),
                     str(seq_per_iso), "", "", "", "", "",
                     str(3 * pham.minimum_length + 3),
                     str(3 * pham.maximum_length + 3),
                     str(3 * pham.average_length + 3)]

        # Deal with Roary rows and strain genes
        for name in genomes.keys():
            if name in pham_genomes:
                geneids = pham_genomes[name]
                roary_row.append(";".join([x.split()[0] for x in geneids]))
                for _ in geneids:
                    strain_rows.append((name, phamid))
            else:
                roary_row.append("")

        roary_rows.append(roary_row)

        fraction = float(len(pham_genomes)) / len(genomes)
        if fraction >= 0.99:
            core.append(phamid)
        elif fraction >= 0.95:
            soft.append(phamid)
        elif fraction >= 0.15:
            shell.append(phamid)
        else:
            cloud.append(phamid)

    summary_file = outdir.joinpath("summary_statistics.txt")
    with open(summary_file, "w") as summary_writer:
        summary = summary_template.format(len(core), len(soft), len(shell),
                                          len(cloud), len(phams))
        summary_writer.write(summary)

    with open(core_file, "w") as core_writer:
        for phamid in core:
            core_writer.write(f"{phamid}\n")

    with open(soft_file, "w") as soft_writer:
        for phamid in soft:
            soft_writer.write(f"{phamid}\n")

    with open(shell_file, "w") as shell_writer:
        for phamid in shell:
            shell_writer.write(f"{phamid}\n")

    with open(cloud_file, "w") as cloud_writer:
        for phamid in cloud:
            cloud_writer.write(f"{phamid}\n")

    roary_file = outdir.joinpath("gene_presence_absence.csv")
    with open(roary_file, "w") as roary_writer:
        header = ROARY_HEADER
        header.extend(genomes.keys())
        roary_writer.write(",".join(header) + "\n")
        for roary_row in roary_rows:
            roary_writer.write(",".join(roary_row) + "\n")

    strain_genes_file = outdir.joinpath("strain_genes.tsv")
    with open(strain_genes_file, "w") as strain_genes_writer:
        for name, phamid in strain_rows:
            strain_genes_writer.write(f"{name}\t{phamid}\n")



"""Classes for efficient storage and utilization of large numbers of
sequences."""

from phamerate.fileio import fasta


class SequenceDB:
    """A convenient interface for wrangling sequences and labels from
    different sources."""
    def __init__(self, sequences=None):
        """Initialize an instance of SequenceDB.

        :param sequences: geneids to add to the database
        :type sequences: list[tuple[str, str]]
        """
        self._translation_geneids: dict[str, list[str]] = dict()
        self._geneid_translations: dict[str, str] = dict()

        if sequences:
            for geneid, translation in sequences:
                self.add_gene(geneid, translation)

    def add_gene(self, geneid, translation):
        """Add a gene to the database.

        :param geneid: the gene's identifier
        :type geneid: str
        :param translation: the gene's protein sequence
        :type translation: str
        """
        if geneid in self._geneid_translations:
            raise ValueError(f"geneid '{geneid}' already in database")

        self._geneid_translations[geneid] = translation

        if translation in self._translation_geneids:
            self._translation_geneids[translation].append(geneid)
        else:
            self._translation_geneids[translation] = [geneid]

    @property
    def geneids(self):
        """Provide access to the geneids in this database.

        :return: geneids
        """
        return self._geneid_translations.keys()

    @property
    def translations(self):
        """Provide access to the translations in this database.

        :return: translations
        """
        return self._translation_geneids.keys()

    @property
    def nr_genes(self):
        """Alternative to __iter__() that allows iterating over
        non-redundant sequences rather than all sequences."""
        for translation, geneids in self._translation_geneids.items():
            yield geneids[0], translation

    def get_translation_geneids(self, translation):
        """Lookup the indicated translation and return its associated
        geneids.

        :param translation: the translation to lookup
        :type translation: str
        :return: geneids
        """
        return self._translation_geneids[translation]

    def get_geneid_translation(self, geneid):
        """Lookup the indicated geneid and return its translation.

        :param geneid: the geneid to lookup
        :type geneid: str
        :return: translation
        :rtype: str
        :raise: KeyError
        """
        return self._geneid_translations[geneid]

    def load(self, filepath, reset=False):
        """Load sequences from a FASTA file.

        :param filepath: the file to load sequences from
        :type filepath: pathlib.Path
        :param reset: clear existing sequences first
        :type reset: bool
        """
        if reset:
            self.__init__()
        for geneid, translation in fasta.read_fasta(filepath):
            self.add_gene(geneid, translation)

    def write(self, filepath, nr_only=False):
        """Dump sequences to a FASTA file.

        :param filepath: path to the file to write_fasta sequences to
        :type filepath: pathlib.Path
        :param nr_only: only write_fasta non-redundant sequences to the file
        :type nr_only: bool
        :return: filepath
        """
        sequences = list()
        if nr_only:
            for geneid, translation in self.nr_genes:
                sequences.append((geneid, translation))
        else:
            for geneid, translation in self:
                sequences.append((geneid, translation))

        return fasta.write_fasta(sequences, filepath)

    def __iter__(self):
        for geneid, translation in self._geneid_translations.items():
            yield geneid, translation

    def __len__(self):
        return len(self._geneid_translations)


class Pham(SequenceDB):
    """A class derived from SequenceDB that is better suited for the
    kinds of operations one might wish to do with smaller groups of
    related sequences.
    """
    def __init__(self, sequences=None):
        super().__init__(sequences)

    def merge(self, other):
        """Merge phams from multiple rounds of clustering.

        :param other: another pham to be merged into this one
        :type other: Pham
        """
        if not isinstance(other, Pham):
            raise TypeError(f"cannot merge Pham with {type(other)}")

        for geneid, translation in other:
            try:
                self.add_gene(geneid, translation)
            except ValueError:
                pass    # Tried adding a gene that was already in the pham

    @property
    def minimum_length(self):
        """Return the length of the shortest sequence in the phamily."""
        lengths = [len(x) for x in self.translations]
        return min(lengths)

    @property
    def average_length(self):
        """Return the length of the average sequence in the phamily."""
        lengths = [len(x) for x in self.translations]
        return float(sum(lengths))/len(lengths)

    @property
    def maximum_length(self):
        """Return the length of the longest sequence in the phamily."""
        lengths = [len(x) for x in self.translations]
        return max(lengths)

    @property
    def n_eff(self):
        """Return the effective size of the pham (number of nr genes)."""
        return len(self.translations)

    @property
    def is_orpham(self):
        """Check whether this pham is an orpham (only 1 gene)."""
        return len(self) == 1

    def update_msa(self, filepath):
        """Introduce duplicate sequences to the MSA in FASTA format.

        :param filepath: path to the MSA file to update
        :type filepath: pathlib.Path
        """
        nr_sequences = fasta.read_fasta(filepath)

        sequences = list()
        for nr_geneid, align in nr_sequences:
            geneids = self.get_translation_geneids(align.replace("-", ""))
            for geneid in geneids:
                sequences.append((geneid, align))

        return fasta.write_fasta(sequences, filepath)

    def __lt__(self, other):
        if not isinstance(other, Pham):
            raise TypeError(f"cannot compare Pham with {type(other)}")

        return len(self) < len(other)



"""Assort phage protein sequences into phamilies of homologs using MMseqs2."""

import pathlib
import shutil
import sys
import tempfile

from phamerate.SequenceDB import Pham, SequenceDB
from phamerate import cli, clustalo, fileio, mmseqs
from phamerate import multiprocess as mp
from phamerate.pangenome import analyze_pangenome


def create_database(input_files):
    """Parse the input file(s) into a SequenceDB object.

    :param input_files: the input file(s) to parse sequences from
    :type input_files: list[pathlib.Path]
    :return: database
    """
    database = SequenceDB()
    for input_file in input_files:
        fmt = fileio.sniff_format(input_file)
        if fmt == "fasta":
            sequences = fileio.read_fasta(input_file)
        elif fmt == "genbank":
            sequences = fileio.read_genbank(input_file)
        else:
            print(f"Unknown file format for {str(input_file)} - skipping")
            continue

        for geneid, translation in sequences:
            try:
                database.add_gene(geneid, translation.upper())
            except ValueError as err:
                print(err)

    return database


def merge_seq_hmm_phams(hmm_phams, seq_phams, database):
    """Merge sequence-sequence and profile-consensus phams.

    :param hmm_phams: profile-consensus clusters
    :type hmm_phams: dict
    :param seq_phams: sequence-sequence clusters
    :type seq_phams: dict
    :param database: the database whose sequences were clustered
    :type database: SequenceDB
    :return: merged_phams
    """
    phams, lookup = list(), dict()
    for key, translations in seq_phams.items():
        pham = Pham()
        for translation in translations:
            lookup[translation] = key
            for geneid in database.get_translation_geneids(translation):
                pham.add_gene(geneid, translation)
        phams.append(pham)

    if not hmm_phams:
        return sorted(phams, reverse=True)

    merged_phams = list()
    for key, translations in hmm_phams.items():
        pham = Pham()
        for translation in translations:
            pre_pham_key = lookup[translation]
            pham.merge(phams[pre_pham_key])
        merged_phams.append(pham)

    return sorted(merged_phams, reverse=True)


def assemble_phams(db, seq_params, hmm_params=None, cpus=1,
                   verbose=False, debug=False):
    """Perform pham assembly on the indicated database.

    If `hmm_params` is `None`, only sequence-sequence clustering will be
    done.

    :param db: the database whose sequences need to be assorted into
        gene phamilies
    :type db: SequenceDB
    :param seq_params: sequence-sequence clustering parameters
    :type seq_params: dict
    :param hmm_params: profile-consensus clustering parameters
    :type hmm_params: dict
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param verbose: print verbose output to console
    :type verbose: bool
    :param debug: run_clustalo MMseqs2 commands in debug mode
    :type debug: bool
    :return: phams
    """
    # Set up temporary directory and name our temporary files
    tmp_dir = pathlib.Path(tempfile.mkdtemp(prefix="phamerate-"))

    nr_fasta = tmp_dir.joinpath("nr_genes.fasta")

    seq_db = tmp_dir.joinpath("sequenceDB")
    clu_db = tmp_dir.joinpath("seqClusterDB")
    sf_db1 = tmp_dir.joinpath("seqSeqfileDB")
    seq_out = tmp_dir.joinpath("seqPhams.txt")

    pro_db = tmp_dir.joinpath("profileDB")
    con_db = tmp_dir.joinpath("consensusDB")
    aln_db = tmp_dir.joinpath("alignDB")
    res_db = tmp_dir.joinpath("hmmClusterDB")
    sf_db2 = tmp_dir.joinpath("hmmSeqfileDB")
    hmm_out = tmp_dir.joinpath("hmmPhams.txt")

    try:
        if verbose:
            print("Creating MMseqs2 database...")

        db.write(nr_fasta, nr_only=True)
        mmseqs.createdb(nr_fasta, seq_db, debug=debug)

        if verbose:
            print("Performing sequence-sequence clustering...")

        # Unpack sequence-sequence clustering parameters, then run_clustalo
        i, c = seq_params["identity"], seq_params["coverage"]
        e, s = seq_params["evalue"], seq_params["sensitivity"]
        m, n = seq_params["cluster_mode"], seq_params["cluster_steps"]
        mmseqs.cluster(seq_db, clu_db, tmp_dir, identity=i, coverage=c,
                       e_value=e, sens=s, cluster_mode=m, cluster_steps=n,
                       cpus=cpus, debug=debug)

        if verbose:
            print("Parsing sequence-sequence phams...")

        mmseqs.createseqfiledb(seq_db, clu_db, sf_db1, cpus=cpus,
                               debug=debug)
        mmseqs.result2flat(seq_db, seq_db, sf_db1, seq_out, debug=debug)

        seq_phams = mmseqs.parse_output(seq_out)

        if hmm_params:
            if verbose:
                print("Building profiles from sequence-sequence phams...")

            mmseqs.result2profile(seq_db, clu_db, pro_db, cpus=cpus,
                                  debug=debug)
            if verbose:
                print("Extracting consensus sequences from profiles...")

            mmseqs.profile2consensus(pro_db, con_db, cpus=cpus,
                                     debug=debug)

            if verbose:
                print("Performing profile-consensus clustering...")

            i, c = hmm_params["identity"], hmm_params["coverage"]
            e, s = hmm_params["evalue"], hmm_params["sensitivity"]
            m, n = hmm_params["cluster_mode"], hmm_params["cluster_steps"]
            mmseqs.search(pro_db, con_db, aln_db, tmp_dir, identity=i,
                          coverage=c, e_value=e, sens=s, cluster_steps=n,
                          cpus=cpus, debug=debug)

            mmseqs.clust(con_db, aln_db, res_db, m, cpus=cpus, debug=debug)

            if verbose:
                print("Parsing profile-consensus phams...")

            mmseqs.createseqfiledb(seq_db, res_db, sf_db2, cpus=cpus,
                                   debug=debug)
            mmseqs.result2flat(pro_db, con_db, sf_db2, hmm_out,
                               debug=debug)

            hmm_phams = mmseqs.parse_output(hmm_out)
        else:
            hmm_phams = None
    except mmseqs.MMseqs2Error as err:
        # Catch MMseqs2 errors, so we can copy temporary files to ~/
        if debug:
            dst = pathlib.Path().home().joinpath(f"{tmp_dir.name}")
            shutil.copytree(src=tmp_dir, dst=dst)
            print(f"Temporary files moved to {dst}")
        # Re-raise the error
        raise err
    finally:
        # Always clear temporary directory regardless of error status
        shutil.rmtree(tmp_dir)

    return merge_seq_hmm_phams(hmm_phams, seq_phams, db)


def write_phams(phams, fasta_dir, nr_only=False):
    """Dump each pham to a FASTA file. Return the path and effective
    size of each pham.

    :param phams: the assembled phams
    :type phams: list[Pham]
    :param fasta_dir: the directory where files should be written
    :type fasta_dir: pathlib.Path
    :param nr_only: only write_fasta non-redundant sequences to file
    :type nr_only: bool
    :return: pham_fastas
    """
    pham_fastas = list()
    for i, pham in enumerate(phams):
        fasta_path = fasta_dir.joinpath(f"pham_{i + 1}.faa")
        pham.write(fasta_path, nr_only=nr_only)
        pham_fastas.append(fasta_path)

    return pham_fastas


def align_phams(phams, fasta_dir, align_dir, cpus=1, verbose=False):
    """Generate pham MSAs with Clustal Omega.

    :param phams: the phams to align
    :type phams: list[Pham]
    :param fasta_dir: path where the fastas should go
    :type fasta_dir: pathlib.Path
    :param align_dir: path where the alignments should go
    :type align_dir: pathlib.Path
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param verbose: print progress bar
    :type verbose: bool
    """
    fasta_paths, align_paths = list(), list()
    jobs = list()
    for i, pham in enumerate(phams):
        fasta_path = fasta_dir.joinpath(f"pham_{i + 1}.faa")
        fasta_paths.append(fasta_path)

        # Not sensible to align phams with only 1 member
        if len(pham) == 1:
            pham.write(fasta_path)
            continue

        align_path = align_dir.joinpath(f"pham_{i + 1}.aln")
        align_paths.append(align_path)

        # If all sequences are the same, they are already aligned
        if pham.n_eff == 1:
            pham.write(fasta_path)
            pham.write(align_path)
            continue

        # All other cases we need to align
        pham.write(fasta_path, nr_only=True)
        jobs.append((fasta_path, align_path))

    results = mp.parallelize(jobs, cpus, clustalo.run_clustalo, verbose)
    for fasta_path, align_path in results:
        i = int(align_path.stem.split("_")[-1]) - 1
        pham = phams[i]

        pham.update_msa(align_path)
        pham.write(fasta_path)

    return fasta_paths, align_paths


def main():
    """Commandline entry point for this module."""
    # Invoke the help menu if phamerate was run_clustalo without args
    if len(sys.argv) == 1:
        sys.argv.append("-h")
    args = cli.parse_args()

    infiles = args.infile
    outdir = args.outdir
    if not outdir.is_dir():
        outdir.mkdir(parents=True)

    skip_hmm = args.skip_hmm
    debug = args.debug
    verbose = args.verbose or debug
    cpus = args.cpus
    do_align = args.align_phams
    do_pangenome = args.pangenome

    # Check for compatible MMseqs2 version - exit if not found
    if not mmseqs.version_ok():
        print(f"\nWrong or missing MMseqs2 dependency - please install version "
              f"{mmseqs.VERSION} and try again\n")
        sys.exit(1)

    # Check for compatible Clustal Omega version - disable align if not found
    if do_align and not clustalo.version_ok():
        print(f"\nWrong or missing Clustal Omega dependency - please install "
              f"version {clustalo.VERSION} to enable pham alignments...\n")
        do_align = False

    # Set up MMseqs2 parameters
    seq_params = {"identity":         args.identity,
                  "coverage":         args.coverage,
                  "evalue":           args.evalue,
                  "sensitivity":      args.sensitivity,
                  "cluster_mode":     args.cluster_mode,
                  "cluster_steps":    args.cluster_steps}

    if skip_hmm:
        hmm_params = None
    else:
        hmm_params = {"identity":      args.hmm_identity,
                      "coverage":      args.hmm_coverage,
                      "evalue":        args.hmm_evalue,
                      "sensitivity":   args.sensitivity,
                      "cluster_mode":  args.hmm_cluster_mode,
                      "cluster_steps": args.hmm_cluster_steps}

    if verbose:
        print("Parsing protein sequences from input files...")

    database = create_database(infiles)

    if verbose:
        print(f"Found {len(database)} translations in {len(infiles)} files...")

    phams = assemble_phams(db=database, seq_params=seq_params,
                           hmm_params=hmm_params, cpus=cpus,
                           verbose=verbose, debug=debug)

    if verbose:
        print(f"Found {len(phams)} phamilies in dataset...")

    fasta_dir = outdir.joinpath("pham_fastas")
    if not fasta_dir.is_dir():
        fasta_dir.mkdir()

    if not do_align:
        write_phams(phams, fasta_dir)
    else:
        if verbose:
            print(f"Computing phamily alignments with Clustal Omega...")

        align_dir = outdir.joinpath("pham_aligns")
        if not align_dir.is_dir():
            align_dir.mkdir()

        align_phams(phams, fasta_dir, align_dir, cpus=cpus, verbose=verbose)

    if do_pangenome:
        if len(infiles) == 1:
            print("Cannot perform pangenome analysis on single input file")
        elif len(infiles) > 1:
            if verbose:
                print("Analyzing pangenome...")

            analyze_pangenome(phams, infiles, outdir)

    print("Done!")


if __name__ == "__main__":
    main()
